/* intextern.c : entrees/sorties structurees */

#include "mlvalues.h"
#include "io.h"
#include "fail.h"
#include "memory.h"
#include "mem_misc.h"

#define BIG_ENDIAN_MAGIC_NUMBER 0x8495A6B7
#define LITTLE_ENDIAN_MAGIC_NUMBER (BIG_ENDIAN_MAGIC_NUMBER + 1)

#define NEVERSEEN WHITE
#define SEEN GRAY
#define SHARED BLACK
#define EMITTED BLUE

#define MARKS COLOUR
#define MARK_SEEN WHITE_TO_GRAY
#define MARK_SHARED GRAY_TO_BLACK
#define MARK_EMITTED BLACK_TO_BLUE
#define MARK_NEVERSEEN ANY_TO_WHITE

struct shared_object {
  value obj;
  value val;
};

static unsigned long total_size, shared_cnt;
static struct shared_object * shared_tbl, * shared_ptr;
static value extern_root;
static unsigned long * extern_ptr;
static unsigned long extern_pos;
extern char * temp_alloc_ptr;

static void full_cleanup(v)
        value v;
{
  value * p;
  blocksize n;

 again:
  if (IS_BLOCK(v) && MARKS(v) != NEVERSEEN) {
    MARK_NEVERSEEN(v);
    if (TAG(v) < NO_SCAN_TAG) {
      for (p = &FIELD(v, 0), n = SIZE(v); n > 1; n--, p++)
        full_cleanup(*p);
      v = *p;                   /* Recursion terminale */
      goto again;
    }
  }
}

static void unshare(v)
        value v;
{
  value * p;
  blocksize n;

 again:
  if (IS_BLOCK(v)) {
    n = SIZE(v);
    if (n == 0) return;
    if (!IS_ALLOCATED(v)) {
      full_cleanup(extern_root);
      failwith("extern: code object");
    }
    switch(MARKS(v)) {
    case NEVERSEEN:
      total_size += 1 + n;
      MARK_SEEN(v);
      if (TAG(v) < NO_SCAN_TAG) {
        for (p = &FIELD(v, 0); n > 1; n--, p++)
          unshare(*p);
        v = *p;                 /* Recursion terminale */
        goto again;
      }
      break;
    case SEEN:
      shared_cnt++;
      MARK_SHARED(v);
      break;
    case SHARED:
      break;
#ifdef DEBUG
    case EMITTED:
      Zassert(0);
#endif
    }
  }
}

static unsigned long emit(v)
     value v;
{
  blocksize n, i;
  value field0;
  unsigned long res;

  if (IS_INT(v))
    return v;
  n = SIZE(v);
  if (n == 0)
    return (TAG(v) << 2) + 2;

  field0 = FIELD(v, 0);

  switch (MARKS(v)) {
#ifdef DEBUG
  case NEVERSEEN:
    Zassert(0);
    break;
#endif
  case SEEN:
    MARK_NEVERSEEN(v);
    break;
  case SHARED:
    MARK_EMITTED(v);
    shared_ptr->obj = v;
    shared_ptr->val = field0;
    shared_ptr++;
    FIELD(v, 0) = extern_pos + 4;
    break;
  case EMITTED:
    return field0;
  }
  *extern_ptr++ = HEADER(v) & ~BLACK;
  *extern_ptr++ = field0;
  for (i = 1; i < n; i++)
    *extern_ptr++ = FIELD(v, i);
  extern_pos += 4;
  res = extern_pos;
  extern_pos += n * sizeof(unsigned long);
  return res;
}

static unsigned long emit_all(root)
     value root;
{
  unsigned long * read_ptr;
  unsigned long res;
  header hdr;
  blocksize n;

  read_ptr = extern_ptr;
  res = emit(root);
  while (read_ptr < extern_ptr) {
    hdr = (header) *read_ptr++;
    n = SIZE_HDR(hdr);
    if (TAG_HDR(hdr) < NO_SCAN_TAG) {
      for ( ; n > 0; n--, read_ptr++)
        *read_ptr = emit(*read_ptr);
    } else {
      read_ptr += n;
    }
  }
  return res;
}

static void cleanup()
{
  struct shared_object * p;

  for (p = shared_tbl; p < shared_ptr; p++) {
    MARK_NEVERSEEN(p->obj);
    FIELD(p->obj, 0) = p->val;
  }
}

value extern_val(chan, v)       /* ML */
     struct channel * chan;
     value v;
{
  unsigned long * extern_block;
  unsigned long res;

  extern_root = v;
  total_size = 0;
  shared_cnt = 0;
  unshare(v);
  if (total_size >= MAX_SIZE) {
    full_cleanup(extern_root);
    failwith("extern: object too big");
  }
  shared_tbl = (struct shared_object *)
    temp_alloc(shared_cnt * sizeof(struct shared_object));
  extern_block = (unsigned long *)
    temp_alloc(total_size * sizeof(unsigned long));
  if (shared_tbl == 0 || extern_block == 0) {
    full_cleanup(extern_root);
    system_failwith_out_of_memory();
  }
  shared_ptr = shared_tbl;
  extern_ptr = extern_block;
  extern_pos = 0;
#ifdef BIG_ENDIAN
  putw(chan, BIG_ENDIAN_MAGIC_NUMBER);
#else
  putw(chan, LITTLE_ENDIAN_MAGIC_NUMBER);
#endif
  putw(chan, total_size);
  res = emit_all(v);
  cleanup();
  temp_free();
  if (total_size == 0)
    putw(chan, res);
  else
    output(chan, (value) extern_block, MLINT(0),
           MLINT(total_size * sizeof(unsigned long)));
  return atom(0);
}

void adjust_pointers(start, size)
     value start;
     blocksize size;
{
  blocksize size_in_bytes, n;
  address p, q;
  header hdr;
  value v;

  size_in_bytes = size * sizeof(value);
  for (p = ADDRESS(start), q = p + size; p < q; ) {
    hdr = (header) *p++;
    n = SIZE_HDR(hdr);
    if (TAG_HDR(hdr) >= NO_SCAN_TAG)
      p += n;
    else
      for( ; n > 0; n--, p++) {
        v = *p;
          switch(v & 3) {
          case 0:                 /* 0 -> bloc represente par son offset */
            Zassert(v >= 0 && v < size_in_bytes && (v & 3) == 0);
            *p = start - 4 + v;
            break;
          case 2:                 /* 2 -> atome */
            v = v >> 2;
            Zassert(v >= 0 && v < 256);
            *p = atom(v);
            break;
          default:                /* 1 ou 3 -> entier */
            break;
          }
      }
  }
}

#define REVERSE_WORD(w) {                                                     \
  char * _p;                                                                  \
  int _a;                                                                     \
  _p = (char *) (w);                                                          \
  _a = _p[0];                                                                 \
  _p[0] = _p[3];                                                              \
  _p[3] = _a;                                                                 \
  _a = _p[1];                                                                 \
  _p[1] = _p[2];                                                              \
  _p[2] = _a;                                                                 \
}

#define REVERSE_DOUBLE(d) {                                                   \
  char * _p;                                                                  \
  int _a;                                                                     \
  _p = (char *) (d);                                                          \
  _a = _p[0];                                                                 \
  _p[0] = _p[7];                                                              \
  _p[7] = _a;                                                                 \
  _a = _p[1];                                                                 \
  _p[1] = _p[6];                                                              \
  _p[6] = _a;                                                                 \
  _a = _p[2];                                                                 \
  _p[2] = _p[5];                                                              \
  _p[5] = _a;                                                                 \
  _a = _p[3];                                                                 \
  _p[3] = _p[4];                                                              \
  _p[4] = _a;                                                                 \
}


void rev_adjust_pointers(start, size)
     value start;
     blocksize size;
{
  blocksize size_in_bytes, n;
  address p, q;
  header hdr;
  value v;

  size_in_bytes = size * sizeof(value);
  for (p = ADDRESS(start), q = p + size; p < q; ) {
    REVERSE_WORD(p)
    hdr = (header) *p++;
    n = SIZE_HDR(hdr);
    switch(TAG_HDR(hdr)) {
    case TAG_STRING:
      p += n;
      break;
    case TAG_FLOAT:
      REVERSE_DOUBLE(p);
      p += n;
      break;
    case TAG_BITVECT:
      for( ; n > 0; n--, p++) {
        REVERSE_WORD(p);
      }
      break; 
    default:
      for( ; n > 0; n--, p++) {
        REVERSE_WORD(p);
        v = *p;
          switch(v & 3) {
          case 0:                 /* 0 -> bloc represente par son offset */
            Zassert(v >= 0 && v < size_in_bytes && (v & 3) == 0);
            *p = start - 4 + v;
            break;
          case 2:                 /* 2 -> atome */
            v = v >> 2;
            Zassert(v >= 0 && v < 256);
            *p = atom(v);
            break;
          default:                /* 1 ou 3 -> entier */
            break;
          }
      }
    }
  }
}

value intern_val(chan)          /* ML */
     struct channel * chan;
{
  blocksize size, size_in_bytes;
  value res;
  unsigned long magic;

  magic =  getw(chan);
  if (magic != BIG_ENDIAN_MAGIC_NUMBER && magic != LITTLE_ENDIAN_MAGIC_NUMBER)
    failwith("intern: bad object");
  size = getw(chan);
  if (size == 0) {
    res = getw(chan);
    if (IS_INT(res))
      return res;
    else
      return atom(res >> 2);
  }
  Zassert(size >= 2);
  size_in_bytes = size * sizeof(value);
  res = alloc(size - 1, TAG_BITVECT);
  if (input(chan, res, MLINT(-4), MLINT(size_in_bytes)) != MLINT(size_in_bytes))
    failwith("intern: truncated object");
#ifdef BIG_ENDIAN
  if (magic == LITTLE_ENDIAN_MAGIC_NUMBER)
#else
  if (magic == BIG_ENDIAN_MAGIC_NUMBER)
#endif
    rev_adjust_pointers(res, size);
  else
    adjust_pointers(res, size);
  return res;
}
  
